--[[
	Copyright (C) 2016  Fuhrbolg - Hyjal

	This program is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program.  If not, see <http://www.gnu.org/licenses/>

	Author's note: Under no circumstances should this Addon be used under any situation.
	It is dangerous and will cause significant harm to your computer. You have been warned.
--]]

--[[
 - 	This file houses all database functionality for battleground match 
 -	history so we can track character specific stats over the course
 -	of a player's career. It will contain the normal BG stats like
 -	killing blows, deaths, damage done, etc. We will be nice and summarize
 -	the player's rank over the course of many BGs.
 -
 -	Encapsulates our database layer. This API will provide the various
 -	save / retrieve operations needed to move data around.
 -
--]]
MatchHistoryService = {}

--[[
 -	Sets up our database API.
--]]
function MatchHistoryService:Init(records)
	self.records = records
	return self
end

--[[
 -	Inserts a MatchHistoryRecord into our database and purges older
 -	records when the table exceeds our limit set in the configuration file.
 -
 -	@param MatchHistoryRecord
--]]
function MatchHistoryService:Save(matchHistoryRecord)
	table.insert(self.records, matchHistoryRecord)
	while table.getn(self.records) > MAX_BG_HISTORY do
		local record = table.remove(self.records, 1)
	end
end

--[[
 -	Getter API that returns table of all MatchHistoryRecords we've recorded.
 -
 -	@return table List of MatchHistoryRecords
--]]
function MatchHistoryService:GetAll()
	return self.records
end

--[[
 -	Purges all MatchHistoryRecords. Can be used by the Options panel to clear data.
--]]
function MatchHistoryService:DeleteAll()
	table.wipe(self.records)
end

--[[
 -	Returns an average value of a numeric filed in a table of tables, using category as the
 -	indexed field of a table within the main table that we are concerned with.
 -	
 - 	@param string
 - 	@return number
--]]
function MatchHistoryService:GetAverage(category)
	local average = 0
	for _, record in pairs(self.records) do
		average = average + record[category]
	end
	return round(average / tablelength(self.records), 2)
end

--[[
 -	Simply adds up the number values mapped to the field specified by category in this table.
 -
 - 	@param string
 - 	@return number
--]]
function MatchHistoryService:GetTotal(category)
	local total = 0
	for _, record in pairs(self.records) do
		if type(record[category]) == "boolean" then
			if record[category] == true then
				total = total + 1
			end
		else
			total = total + record[category]
		end
	end
	return total
end

--[[
 -	Iterate through all records and find the record key that has the highest numeric value.
 -	
 - 	@param string
 - 	@return number
--]]
function MatchHistoryService:GetHighestValue(category)
	local highest = 0
	for _, record in pairs(self.records) do
		local value = record[category]
		if value > highest then
			highest = value
		end
	end
	return highest
end

--[[
 -	Compares two numeric values identified by category1 and category2
 -	producing a ratio between the two values.
 -
 -	@param function f1 Returns first value from record
 -	@param function f2 Returns second value from record
 -	@return double
--]]
function MatchHistoryService:GetRatio(f1, f2)
	local val1 = 0
	local val2 = 0
	for _, record in pairs(self.records) do
		val1 = val1 + f1(record)
		val2 = val2 + f2(record)
	end
	if val2 == 0 then
		return 100
	end
	return val1 / (val1 + val2) * 100
end

--[[
 -	Returns an array of tables This is intended
 -	to be used by the chart functions for mapping data to a line graph.
 -
 -	Rankings can be either rank|damageRank|healingRank
 -	
 -	Example return table:
 -	{{0,5},{1,4},{2,5},{3,3}}	
 -
 -	@param string category
 -	@return table
--]]
function MatchHistoryService:GetTrendData(category)
	local sorted = {}
	local index = 0
	for _, record in spairs(self.records, function(t,a,b) 
		return t[b].time > t[a].time
	end) do
		local value = record[category]
		table.insert(sorted, { index, value } )
		index = index + 1
	end
	return sorted
end

--[[
 -	Returns an array of rankings based on the category provided sorted by
 -	battleground time. This is keyed with a zero based index.
 -
 -	Example return table:
 - 	sorted[0] = 35
 -	sorted[1] = 21
 -	sorted[2] = 10
 -	sorted[3] = 10
 -	sorted[4] = 10
 -	sorted[5] = 15
 -
 -	@param string category
 -	@return table
--]]
function MatchHistoryService:GetRankPercentBreakdownData(category)
	local sorted = { [0] = 0, [1] = 0, [2] = 0, [3] = 0, [4] = 0, [5] = 0}
	local total = tablelength(self.records)
	local multiplier = 100 / tablelength(self.records)
	local index = 0
	for _, record in spairs(self.records, function(t,a,b) 
		return t[b].time > t[a].time
	end) do
		local key = record[category] > 5 and 0 or record[category]
		sorted[key] = sorted[key] + 1
		index = index + 1
	end
	for key, value in pairs(sorted) do
		sorted[key] = value * multiplier
	end
	return sorted
end

--[[
 -	This class represents a single batle ground result. It contains all of the
 - 	player stats like damage, healing, killing blows, etc. in addition to some
 -	meta information like BG map, whether the player won, etc.
--]]
MatchHistoryRecord = {}
MatchHistoryRecord.__index = MatchHistoryRecord

--[[
 -	Constructs a new MatchHistoryRecord instance from the
 -	player object provided. 
 -
 -	Usage: local record = MatchHistoryRecord.new(player)
 -
 - 	Note: This was copied from the Winning addon by Salovia
 - 	https://mods.curse.com/addons/wow/winning
 -
 -	@param table Our Player object
 -	@param table List of Player objects
 -	@return MatchHistoryRecord
--]]
MatchHistoryRecord.New = function(player, players)
	if not player or not players then return end
	
	-- Find total damage and healing for both teams. 
	local enemyDamage = 0
	local enemyHealing = 0
	local teamDamage = 0
	local teamHealing = 0
	local teamKills = 0
	local teamDeaths = 0
	local enemyKills = 0
	local enemyDeaths = 0
	
	local teamSize = 0
	local avgTeamKills = 0
	local avgTeamDeaths = 0
	local avgTeamDamage = 0
	local avgTeamHealing = 0
	
	for _, p in pairs(players) do	
		if p.enemy then
			enemyDamage = enemyDamage + p.damage
			enemyHealing = enemyHealing + p.healing
			enemyKills = enemyKills + p.kills
			enemyDeaths = enemyDeaths + p.deaths
		else
			teamDamage = teamDamage + p.damage
			teamHealing = teamHealing + p.healing
			teamKills = teamKills + p.kills
			teamDeaths = teamDeaths + p.deaths
			if player.GUID ~= p.GUID then
				teamSize = teamSize + 1
				avgTeamKills = avgTeamKills + p.kills
				avgTeamDamage = avgTeamDamage + p.damage
				avgTeamHealing = avgTeamHealing + p.healing
				avgTeamDeaths = avgTeamDeaths + p.deaths
			end
		end
	end
	
	-- NOTE: This is very important. In the unlikely event that a player reloads
	-- his UI after a BG match is over MongoMon will attempt to save the data (again).
	-- However, the GUID values for players will be null as they are scraped from the 
	-- combat log. So since every value is nil the teamSize will be 0. And the below
	-- calculations will result in nil values. We could avoid this by simply checking
	-- the player names, however duplicate stats would still be recorded. The end
	-- of battle ground event signal will fire again when the UI reloads. So we're
	-- leaving this strangeness here to handle that scenario. Another approach is
	-- to create a token for each battleground and check whether it exists in the database
	-- before saving. Too boring.
	if teamSize == 0 then return end	

	avgTeamKills = avgTeamKills / teamSize
	avgTeamDeaths = avgTeamDeaths / teamSize
	avgTeamDamage = avgTeamDamage / teamSize
	avgTeamHealing = avgTeamHealing / teamSize
	
	-- Compute player's percentage of team's output
	local playerPercentDamageOfTeam = string.format("%.2f", player.damage / teamDamage * 100)
	local playerPercentHealingOfTeam = string.format("%.2f", player.healing / teamHealing * 100)	
	local playerPercentKillsOfTeam = string.format("%.2f", player.kills / teamKills * 100)	
	
	-- Create our new and improved MatchHistoryRecord
	local self = {}
	setmetatable(self, MatchHistoryRecord)
	
	-- Add the main BG stats from the player object
	self.name = player.name
	self.kills = player.kills
	self.deaths = player.deaths
	self.damage = player.damage
	self.healing = player.healing
	self.class = player.class
	self.spec = player.spec
	self.rank = player.rank
	self.damageRank = player.damageRank
	self.healingRank = player.healingRank
	self.faction = player.faction
	self.isHealer = player.isHealer
	
	-- Add some extra values to our player object
	self.time = time()
	self.enemyDamage = enemyDamage
	self.enemyHealing = enemyHealing
	self.teamDamage = teamDamage
	self.teamHealing = teamHealing
	self.teamKills = teamKills
	self.teamDeaths = teamDeaths
	self.enemyKills = enemyKills
	self.enemyDeaths = enemyDeaths
	self.avgTeamKills = avgTeamKills
	self.avgTeamDeaths = avgTeamDeaths
	self.avgTeamDamage = avgTeamDamage
	self.avgTeamHealing = avgTeamHealing
	self.playerPercentDamageOfTeam = playerPercentDamageOfTeam
	self.playerPercentHealingOfTeam = playerPercentHealingOfTeam
	self.playerPercentKillsOfTeam = playerPercentKillsOfTeam
	
	local battlefieldWinner = GetBattlefieldWinner() -- Returns 0 for horder 1 for alliance, nil for not over
	if battlefieldWinner ~= nil then
		self.didPlayerWin = FACTIONS[battlefieldWinner] == player.faction
	else
		self.didPlayerWin = false
	end
	
	SetMapToCurrentZone()
	self.mapId = GetCurrentMapAreaID()
	self.season = GetCurrentArenaSeason()
	self.isRated = IsRatedBattleground()	
	
    return self
end